#!/usr/bin/env python3

# Watchdog used to check Hard Drives and report on their status

import sys
import os
import socket
import subprocess
import io
import re
from collections import OrderedDict


class WatchAPHDD(object):
    DEBUG = False

    def __init__(self, watchdog_csv_dir: str, watchdog_tmp_file: str, watchdog_csv_file: str):
        # machinelocalwatchdog folder and files to use for DM report
        self.watchdog_csv_dir = watchdog_csv_dir
        self.watchdog_tmp_file = watchdog_tmp_file
        self.watchdog_csv_file = watchdog_csv_file

        # Get hostname
        self.hostname = socket.gethostname()

        # Store raw devices
        self.raw_devices = []

        # Store processed results
        self.device_results = {}

    # Use lsblk to determine what block devices are in the system
    def get_block_devices(self) -> str:
        command = "lsblk -S -n -o name,type"

        # Example lsblk output
        # sda  disk
        # sdb  disk
        # sdc  disk

        if self.DEBUG:
            print("\nwatchaphdd:get_block_devices Command: " + command)

        try:
            result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('watchaphdd:get_block_devices subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('watchaphdd:get_block_devices subprocess.CalledProcessError')

        return result

    # Use lsblk to determine what block devices are in the system
    def process_block_devices(self) -> bool:
        try:
            result = self.get_block_devices()
        except Exception as e:
            print("watchaphdd.process_block_devices: Exception: " + str(e))
            return False

        raw_devices_list = result.strip().split("\n")
        if self.DEBUG:
            print("\nRaw Block Devices:")
            print(raw_devices_list)

        raw_devices = {}
        for device_data in raw_devices_list:
            raw_devices[device_data.strip().split()[0]] = device_data.strip().split()[1]

        # Make sure devices aren't all null
        device_found = False
        for device, type in raw_devices.items():
            if device:
                device_found = True

        if not device_found:
            return False

        # Make sure devices exclude anything that's not "disk" type
        for device in list(raw_devices):
            device_type = raw_devices[device]
            if device_type != "disk":
                if self.DEBUG:
                    print("Found device: " + device + " of type: " + device_type + " which is not type disk: Removing")
                del raw_devices[device]

        # Make sure more than 1 device is found
        if len(raw_devices) < 2:
            return False
        else:
            self.raw_devices = raw_devices

        if self.DEBUG:
            print("\nStripped Raw Block Devices:")
            print(self.raw_devices)

        return True

    # Get specific block device type for device
    def get_block_device_type(self, device: str) -> str:
        command = "cat /sys/block/" + device + "/queue/rotational"

        if self.DEBUG:
            print("watchaphdd:get_block_device_type " + command)
        try:
            result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('main.get_block_device_type exception thrown subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('main.get_block_device_type exception thrown subprocess.CalledProcessError')

        return result

    # Check to see if the device is rotational or not
    def get_types_for_block_devices(self) -> bool:
        for device in list(filter(None, self.raw_devices)):
            # Determine whether the device is an SSD or an HDD by probing the rotational parameter
            if self.DEBUG:
                print("\nwatchaphdd:get_types_for_block_devices Getting device type information on /dev/" + device)

            try:
                result = self.get_block_device_type(device)
            except Exception as e:
                print("watchaphdd.get_types_for_block_devices: Exception: " + str(e))
                return False

            device_type_value = result.split("\n")[0]

            if not device_type_value:
                print("\nwatchaphdd:get_types_for_block_devices Unable to determine HDD vs SSD for device: " + device)
                return False

            if (device_type_value == "1"):
                device_type = "HDD"
                if self.DEBUG:
                    print("watchaphdd:get_types_for_block_devices Found HDD rotational device")
            else:
                device_type = "SSD"
                if self.DEBUG:
                    print("watchaphdd:get_types_for_block_devices Found SSD device")

            self.device_results[device] = {"type": device_type}

        return True

    # Get the name of the device Bus ID using lsscsi
    def get_lsscsi_device_bus_data(self) -> str:
        command = "lsscsi --verbose | egrep " + r'"^\["' + r" | awk '{ print $1 " + '" "' + r" $NF }'"
        if self.DEBUG:
            print("\nwatchaphdd:get_lsscsi_device_bus Running lsscsi command: " + command)

        try:
            result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('main.get_lsscsi_device_bus exception thrown subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('main.get_lsscsi_device_bus exception thrown subprocess.CalledProcessError')

        return result

    def process_lsscsi_device_bus_data(self) -> bool:
        try:
            result = self.get_lsscsi_device_bus_data()
        except:
            return False

        device_list = result.split("\n")

        # Example Device List - all kinds of crazy numbering and bogus values for device names are possible
        # We make sure device names are valid and all 4 bus numbers are reported
        # [0:0:0:0] -
        # [1:0:1:0] /dev/sda
        # [2:0:2:0] /dev/sdb
        # [3:0:3:0] /dev/sdc
        # [0:0:4:0] /dev/sdd
        # [0:0:5:0] /dev/sde
        # [0:0:6:0] /dev/sdf
        # [0:0:7:0] -

        for device in list(filter(None, device_list)):

            # Make sure the bus format is in expected notation, otherwise we return False and error out
            # Format is this: [0:0:0:0]
            raw_bus = device.split(" ")[0]
            if self.DEBUG:
                print("Raw Bus: " + raw_bus)

            # Example Bus Output
            # "0:0:0:0"
            # "1:0:0:0"
            # "2:0:0:0"
            # "3:0:0:0"
            # "0:0:14:0"
            # "0:0:0:0"
            # "0:4:3:0"

            # 1:0:0:0
            bus_pattern = re.compile('^\[[0-9]+:[0-9]+:[0-9]+:[0-9]+\]$')
            match_bus = bus_pattern.search(raw_bus)

            if not raw_bus or not match_bus:
                if self.DEBUG:
                    print("Bus missing or malformed: " + raw_bus)
                return False

            # /dev/sda
            raw_device = device.split(" ")[1]
            if self.DEBUG:
                print("Raw Device: " + raw_device)

            raw_device_pattern = re.compile('\/dev\/[a-zA-Z0-9]+')
            raw_device_match = raw_device_pattern.search(raw_device)

            if raw_device_match:
                # ['', 'dev', 'sda']
                device = raw_device.split("/")[2]
                if self.DEBUG:
                    print("Device: " + device)

                device_pattern = re.compile('[a-zA-Z0-9]+')
                device_match = device_pattern.search(device)

                if device_match:
                    if device.strip() in self.device_results:
                        for key, value in self.device_results.items():
                            if key.strip() == device.strip():
                                if self.DEBUG:
                                    print("Found matching device key: " + key.strip() + " in self.device_results")
                                self.device_results[device]["bus"] = raw_bus
                                # Used to sort by bus when reporting/displaying output
                                self.device_results[device]["numeric_bus"] = "".join(raw_bus.split(":"))
                else:
                    if self.DEBUG:
                        print("Skipping device: " + device)
            else:
                if self.DEBUG:
                    print("Skipping raw device: " + raw_device)

        return True

    # Get device SMART health
    def get_smart_status(self, device: str) -> bool:
        command = "smartctl -H /dev/" + device + " | grep PASSED"
        if self.DEBUG:
            print("\nwatchaphdd:get_smartctl_info Running smartctl command: " + command)

        try:
            result_smart = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired as e:
            print("watchaphdd.get_smart_status: Exception: " + str(e))
            return False
        except subprocess.CalledProcessError as e:
            print("watchaphdd.get_smart_status: Exception: " + str(e))
            return False

        result_smart_stripped = result_smart.strip()
        pattern_smart_passed = re.compile('PASSED')
        match_smart_passed = pattern_smart_passed.search(result_smart_stripped)

        if match_smart_passed:
            return True
        else:
            return False

    def get_smart_details(self, device: str) -> str:
        command = "smartctl -i /dev/" + device + " | egrep 'Model|Serial|Firmware'"
        if self.DEBUG:
            print("\nwatchaphdd:get_smartctl_info Running smartctl command: " + command)

        try:
            result_details = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('main.get_smart_details exception thrown subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('main.get_smart_details exception thrown subprocess.CalledProcessError')

        return result_details

    # use smartctl to get device health and device detailed information like firmware version, etc.
    def get_smartctl_info(self) -> bool:
        for device in list(filter(None, self.raw_devices)):
            device_model = ""
            device_serial = ""
            device_firmware = ""
            smart_health_check_passed = False
            smart_result_details = ""
            smart_error = False

            if self.get_smart_status(device):
                smart_health_check_passed = True
            else:
                smart_health_check_passed = False
                smart_error = True

            # We only do a complete property check on the hard drive if it passes SMART health test
            # since we don't want to lock up the machine if the drive is dying
            if not smart_error:
                try:
                    smart_result_details = self.get_smart_details(device)
                except:
                    smart_error = True

            # If the device is in error state we mark all properties as being in error because we do not want to parse
            # details on a bad drive as it may lock up the system
            if smart_error or not smart_health_check_passed:
                device_model = "Error"
                device_serial = "Error"
                device_firmware = "Error"

                if self.DEBUG:
                    print("Device: " + device + " is in SMART Error")

                if device.strip() in self.device_results:
                    for key, value in self.device_results.items():
                        if key.strip() == device.strip():
                            # Switch to Fatal once we're sure no other weird device names show up
                            # self.device_results[device]["status"] = "ErrorFatal"
                            self.device_results[device]["status"] = "Warning"
                            self.device_results[device]["model"] = device_model
                            self.device_results[device]["serial"] = device_serial
                            self.device_results[device]["firmware"] = device_firmware
            else:
                if not smart_result_details:
                    device_model = "Error"
                    device_serial = "Error"
                    device_firmware = "Error"

                    if self.DEBUG:
                        print("Device: " + device + " is in SMART Error because it had no SMART results")

                else:
                    # ['Device Model:     SK hynix SH921 2.5 7MM 480GB', 'Serial Number:    1IE7T000110507B0P', Firmware Version: 10011L00', '']
                    raw_device_details = smart_result_details.split("\n")

                    for details in list(filter(None, raw_device_details)):
                        pattern_model = re.compile('Model')
                        pattern_serial = re.compile('Serial')
                        pattern_firmware = re.compile('Firmware')
                        match_model = pattern_model.search(details)
                        match_serial = pattern_serial.search(details)
                        match_firmware = pattern_firmware.search(details)

                        if match_model:
                            device_model = details.split(":")[1].strip()

                        if match_serial:
                            device_serial = details.split(":")[1].strip()

                        if match_firmware:
                            device_firmware = details.split(":")[1].strip()

                        if not match_model and not match_serial and not match_firmware:
                            device_model = "Error"
                            device_serial = "Error"
                            device_firmware = "Error"

                if device.strip() in self.device_results:
                    for key, value in self.device_results.items():
                        if key.strip() == device.strip():
                            if not smart_health_check_passed:
                                # Switch to Fatal once we're sure no other weird device names show up
                                # self.device_results[device]["status"] = "ErrorFatal"
                                self.device_results[device]["status"] = "Warning"
                            else:
                                self.device_results[device]["status"] = "Ok"

                            self.device_results[device]["model"] = device_model
                            self.device_results[device]["serial"] = device_serial
                            self.device_results[device]["firmware"] = device_firmware
                else:
                    return False

        return True

    # Sort all devices in an ordered dictionary and print out the results
    def print_device_results(self):
        try:
            print("\nDevice Results:")
            print(OrderedDict(sorted(self.device_results.items(), key=lambda x: x[1]['bus'])).items())
        except Exception as e:
            print("watchaphdd.print_device_results: Exception: " + str(e))
            return False

        return True

    # Create DM report out of the information returned
    def build_dm_report(self):
        self.dm_output = io.StringIO()

        # CO4PHY080101813,Ok,smartTst:[1:0:0:0],HDD disk [1:0:0:0] model 'TOSHIBA MG03ACA200' serial 'Y4R1KHM1F' firmware 'FL1H' is Ok
        for device, value in OrderedDict(sorted(self.device_results.items(), key=lambda x: '{0:0>8}'.format(x[1]['numeric_bus']).lower())).items():
            line = [self.hostname,
                    self.device_results[device]["status"], "smartTst:" +
                    self.device_results[device]["bus"],
                    self.device_results[device]["type"] + " disk " +
                    self.device_results[device]["bus"] + " model '" +
                    self.device_results[device]["model"] + "' serial '" +
                    self.device_results[device]["serial"] + "' firmware '" +
                    self.device_results[device]["firmware"] + "' is " +
                    self.device_results[device]["status"]]

            try:
                self.dm_output.write(",".join(line) + '\r\n')
            except Exception as e:
                print("watchaphdd.build_dm_report: Exception: " + str(e))
                return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
            if not len(results) == 4:
                print("watchaphdd.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.watchdog_csv_dir):
            print("watchaphdd.write_dm_report_to_disk: Directory: " + self.watchdog_csv_dir + " not found")
            return False
        else:
            try:
                csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

                with open(csv_file_path, "w") as text_file:
                    text_file.write(self.dm_output.getvalue())

            except Exception as e:
                print("watchaphdd.write_dm_report_to_disk: Exception: " + str(e))
                return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'r') as content_file:
                content = content_file.read()

                content = content.split("\n")
                dm_output = self.dm_output.getvalue().split("\r\n")

                if list(filter(None, content)) == list(filter(None, dm_output)):
                    return True
                else:
                    print("watchaphdd.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                    return False

        else:
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_csv_file)

        # print("\nRenaming: " + tmp_file_path + " to: " + csv_file_path)

        try:
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            print("watchaphdd.atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("watchaphdd.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Execute all the methods in order
    def main(self):
        if not self.get_block_devices():
            print("watchaphdd.main: Could not discover block devices (main.get_block_devices)")
            sys.exit(1)

        if not self.process_block_devices():
            print("watchaphdd.main: Could not process block devices (main.process_block_devices)")
            sys.exit(1)

        if not self.get_types_for_block_devices():
            print("watchaphdd.main: Could not get block device types (main.get_types_for_block_devices)")
            sys.exit(1)

        if not self.process_lsscsi_device_bus_data():
            print("watchaphdd.main: Could not get lsscsi information (main.get_lsscsi_device_bus)")
            sys.exit(1)

        if not self.get_smartctl_info():
            print("watchaphdd.main: Could not get smartctl information (main.get_smartctl_info)")
            sys.exit(1)

        if self.DEBUG:
            if not self.print_device_results():
                print("watchaphdd.main: Could not print device results (main.print_device_results)")
                sys.exit(1)

        if not self.build_dm_report():
            print("watchaphdd.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("watchaphdd.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("watchaphdd.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.verify_write_dm_report_to_disk():
            print("watchaphdd.main: Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print("watchaphdd.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        if self.DEBUG:
            if not self.print_dm_report():
                print("watchaphdd.main: Could not print DM report (main.print_dm_report)")
                sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally

if __name__ == '__main__':
    watchdog_csv_dir = "/ap/data/watchdogoutput/csv_sources"
    watchdog_tmp_file = "watchaphdd.tmp"
    watchdog_csv_file = "watchaphdd.csv"

    WatchAPHDD(watchdog_csv_dir, watchdog_tmp_file, watchdog_csv_file).main()
